/* HA.C -- Example LONWORKS(R) Host Application.
 * Copyright (c) 1993, 1994 by Echelon Corporation.
 * All Rights Reserved.
 */

#define     VERSION    "3.0"

/*
 * Date last modified: 11/17/94
 *
 * Main program for an example LONWORKS host application.  The example
 * sends and receives both polls and updates to network variables.  The
 * host application node declares and implements eight network variables.
 * The example also processes incoming network management commands for
 * network variable binding so that network variables may be bound by a
 * network management tool.
 *
 * The example uses a LONWORKS standard network driver to communicate with
 * a LONWORKS network interface such as the Echelon Serial LonTalk Adapter
 * (SLTA), PC LonTalk Adapter (PCLTA), or a Microprocessor Interface Program
 * (MIP)-based network interface.
 *
 * The example implements an application node that will be installed by
 * another network management tool.
 */

/****************************** Include Files *******************************/

#include "ni_msg.h"
#include "ni_mgmt.h"
#include "hauif.h"
#include "applmsg.h"
#include <stdio.h>
#include <stdlib.h>
#include <conio.h>
#include <ctype.h>
const char * lonDeviceName = "LON1";

/*************************** Function Prototypes ****************************/


    // prototypes
void errorExit( void );
void signOn( void );
Bool installProgId( void );
void getArgs( int argc, char * argv[] );

/*
 ****************************************************************************
 *
 *      main().  Display sign-on message, open network driver, and enter
 *               the main scheduler loop.
 *
****************************************************************************
 */

int main( int argc, char * argv[] ) {

    NICode       niError;
    const char * prompt  = "HA> ";

    Bool         promptFlag = TRUE;
    Bool         exitFlag;

    getArgs( argc, argv );     // get command-line argumenta
    signOn( );                 // print sign on message

    // Open and initialize the network interface

    niError = NiInit( lonDeviceName );

    if( niError!= NI_OK ) {
        NiErrorDisplay( "Error while initializing network interface",
            niError );
        return 1;
    }

    if( installProgId( ) ) return 1;
                         // install program ID into network interface
    loadNvConfig( );     // load network variable config table from disk

    do {
        if( promptFlag ) printf( prompt );
        exitFlag = processCmd( prompt );   // handle keyboard input
        promptFlag = processMsg( );        // handle network input
    } while( ! exitFlag );                 // main processing loop

    storeNvConfig( );
    NiClose( );
    printf("Leaving the Host Application program.");
    return 0;
}

/*
 ****************************************************************************
 *
 *      errorExit - handle invocation errors
 *
 ****************************************************************************
 */

void errorExit( void ) {

    static const char flagMessage[] = {
      "\n\n"
      "To invoke the LONWORKS Host Application Demonstration Program, enter:"
      "\n\n"
      "    ha [-Ddevice] [-V]\n\n"
      "    -D selects the network device name, the default is 'LON1'.\n"
      "    -V invokes verbose mode.\n"
    };

    printf( flagMessage );
    exit( 1 );
}

/*
 ****************************************************************************
 *
 *      Sign-on message
 *
 ****************************************************************************
 */

void signOn( void ) {

    extern void helpCmd( void );

    static const char signOnMessage[] = {
      "Welcome to the LONWORKS(R) host application, version %s\n\n"
      "This program acts as an application node\n"
      "Enter one of the following commands by typing the indicated letter:\n\n"
    };

    printf( signOnMessage, VERSION );    // print sign-on
    helpCmd( );                          // print menu
}

/*
 ****************************************************************************
 *
 *      Verbose mode command
 *
 ****************************************************************************
 */

void verbose( void ) {       // Control verbose and report modes

    verboseFlag = getChoice(
        "Display network interface msgs (Y/N) ?",
        'Y', 'N' );
    reportFlag = getChoice(
        "Report incoming appl msgs (Y/N) ?",
        'Y', 'N' );
}

/*
 ****************************************************************************
 *
 *      installProgId - install a program ID into the network interface
 *
 ****************************************************************************
 */

Bool installProgId( void ) {
    struct {
        NM_write_memory_request     writeMemHdr;
        char     progID[ ID_STR_LEN ];           // data to be written
    } writeMemMsg = {
         {  NM_write_memory,      // code
            READ_ONLY_RELATIVE,   // mode
            0x00, 0x0D,           // offset for id_string
            ID_STR_LEN,           // count
            BOTH_CS_RECALC },     // form
            "HostAppl" };         // program ID

    ComplType       completion;
    MsgData         response;
    NICode          niError;
    SendAddrDtl     netIntfcAddr;

    netIntfcAddr.lc.type = LOCAL;     // send it to the network intfc
    niError = NiSendMsgWait(
        REQUEST,
        & netIntfcAddr,         // out_addr
        ( MsgData * ) &writeMemMsg,  // out_data
        sizeof( writeMemMsg ),  // out_length
        FALSE,                    // priority
        FALSE,                    // out_auth
        & completion,             // completion code
        NULL,               // num_responses
        NULL,               // inAddr
        & response,         // inData
        NULL,               // inLength
        NI_REFERENCE_ID_ANY );

    return NiHandleError( niError, completion, response.exp.code,
        "updating program ID" );           // report any errors
}

/*
 ****************************************************************************
 *
 *      getArgs - process command line arguments
 *
 ****************************************************************************
 */

void getArgs( int argc, char * argv[ ] ) {
    const char * pArg;
    char ch;

    while( --argc ) {
        pArg = argv[ argc ];
        if( ( * pArg == '/' ) || ( * pArg == '-') ) {
            ch = * ( ++pArg );
            switch( toupper( ch ) ) {
            case 'D':       // device name
                if( * (++pArg) ) lonDeviceName = pArg;
                    // Device name specified, copy it
                else errorExit( );
                break;

            case 'V':       // verbose mode
                /* Set verbose mode */
                verboseFlag = TRUE;
                break;

            default:        // invalid switch
                errorExit( );
            }              // end switch
        } else errorExit( );        // invalid first character
    }                      // end while
}
