/* LDVINTFC.C -- LONWORKS(R) standard network interface functions
 * Copyright (c) 1993, 1994 by Echelon Corporation.
 * All Rights Reserved.
 *
 * Date last modified: 11/17/94
 *
 * Example Standard LONWORKS network interface functions
 * This layer contains all the DOS dependencies of the host application
 * interface to the network device.
 */

/****************************** Include Files *******************************/

#include <io.h>
#include <fcntl.h>
#include <sys\types.h>
#include <sys\stat.h>
#include <stdio.h>
#ifdef _MSC_VER
    extern int ioctl( int handle, int func, void far * argdx, int argcx );
#endif

#include "ldvintfc.h"
#include "hauif.h"

/*
 ****************************************************************************
 * Network driver interface structure.  This structure is used to pass the
 * read, write, and register function pointers back via the ioctl() call.
 ****************************************************************************
 */

typedef struct {
    Byte    Status;
    LDVCode ( far * readFn ) ( void far * pMsg,  unsigned length );
                 // Read function pointer
    LDVCode ( far * writeFn ) ( void far * pMsg, unsigned length  );
                 // Write function pointer
    void    ( far * registerFn ) ( int handle,
                 void ( interrupt far * callback ) ( void ) );
                 // Register function pointer
} directCalls;

/*
 ****************************************************************************
 * Device table - stores resources associated with each open
 * network interface.
 ****************************************************************************
 */

#define NUM_DEVICES 8

typedef struct {
    int         dosHandle;      // -1 means entry is unused
    directCalls calls;
} device;

static device deviceTable[ NUM_DEVICES ] =
  { { -1 }, { -1 }, { -1 }, { -1 }, { -1 }, { -1 }, { -1 }, { -1 } };

/*
 ****************************************************************************
 * ldv_open() - open network interface device
 ****************************************************************************
 */

LDVCode ldv_open( const char * deviceName, LNI * pHandle ) {
    int         dosHandle;
    LNI         lniHandle;
    device    * pDevice;

    // find a free entry in the device table
    for( lniHandle = 0, pDevice = deviceTable; lniHandle < NUM_DEVICES;
        lniHandle++, pDevice++ ) {
        if( pDevice->dosHandle == -1 ) break;
    }
    if( lniHandle >= NUM_DEVICES ) return LDV_NO_RESOURCES;
            // error - too many opens

    // open the device with DOS
    dosHandle = open( deviceName, O_RDWR | O_BINARY );
    if( dosHandle < 0 ) {          // DOS error
        perror( "Error opening network driver" );
        return LDV_NO_RESOURCES;
    }

    // get direct calls entry points
    if( ioctl( dosHandle, 2, & pDevice->calls, sizeof( directCalls ) )
        < 0 ) {
        perror( "Error initializing network interface" );
        close( dosHandle );
        return LDV_DEVICE_ERR;                  // error calling ioctl
    }
    pDevice->dosHandle = dosHandle;
    * pHandle = lniHandle;
    return LDV_OK;
}

/*
 ****************************************************************************
 * ldv_read() - read buffer from network interface
 ****************************************************************************
 */

LDVCode ldv_read( LNI lniHandle, void * pMsg, short length ) {
    device      * pDevice;

    if( lniHandle < 0 ) return LDV_NOT_OPEN;
    if( lniHandle > NUM_DEVICES ) return LDV_INVALID_DEVICE_ID;
    pDevice = & deviceTable[ lniHandle ];
    if( pDevice->dosHandle == -1 ) return LDV_NOT_OPEN;
    return pDevice->calls.readFn( pMsg, length );
}

/*
 ****************************************************************************
 * ldv_write() - write buffer to network interface
 ****************************************************************************
 */

LDVCode ldv_write( LNI lniHandle, void * pMsg, short length ) {
    device      * pDevice;

    if( lniHandle < 0 ) return LDV_NOT_OPEN;
    if( lniHandle > NUM_DEVICES ) return LDV_INVALID_DEVICE_ID;
    pDevice = & deviceTable[ lniHandle ];
    if( pDevice->dosHandle == -1 ) return LDV_NOT_OPEN;
    return pDevice->calls.writeFn( pMsg, length );
}

/*
 ****************************************************************************
 * ldv_close() - close network interface
 ****************************************************************************
 */

LDVCode ldv_close( LNI lniHandle ) {
    device      * pDevice;

    if( lniHandle < 0 ) return LDV_NOT_OPEN;
    if( lniHandle > NUM_DEVICES ) return LDV_INVALID_DEVICE_ID;
    pDevice = & deviceTable[ lniHandle ];
    if( pDevice->dosHandle == -1 ) return LDV_NOT_OPEN;
    if( close( pDevice->dosHandle ) < 0 ) {
        perror( "Error closing device" );
        return LDV_DEVICE_ERR;
    }
    pDevice->dosHandle = -1;      // show it's free
    return LDV_OK;
}

/*
 ****************************************************************************
 * Network driver error strings.
 ****************************************************************************
 */

static const char * ldvErrorStrings[] = {
    "No error",
    "Network driver not found",
    "Network interface already open",
    "Network interface not open",
    "Network interface hardware error",
    "Invalid device ID",
    "No messages buffered for read",
    "No message buffers available for write",
    "Network interface being initialized"
};

/*
 ****************************************************************************
 * ldvErrorDisplay().  Display a network driver error message.  Display
 * a user specified string first.
 ****************************************************************************
 */

void ldvErrorDisplay( const char * str, LDVCode ldvError ) {
    if( ldvError == LDV_OK ) return;
    errorDisplay( str, ldvErrorStrings[ ldvError ] );
}

/*
 ****************************************************************************
 *
 * Functions to handle byte ordering
 *
 ****************************************************************************
 */

typedef union {               // overlay two bytes on a 16-bit word
    Word int_16;
    struct {
        Byte byte_1;
        Byte byte_2;
    } b;
} union_16;

Word swapBytes( Word int_16 ) {             // swap bytes in 16-bit num
    union_16 u1, u2;
    u1.int_16 = int_16;
    u2.b.byte_2 = u1.b.byte_1;
    u2.b.byte_1 = u1.b.byte_2;
    return u2.int_16;
}

Word makeWord( Byte hi, Byte lo ) {        // make 16-bit word from bytes
    union_16 u;
    u.b.byte_1 = lo;         // Intel order
    u.b.byte_2 = hi;
    return u.int_16;
}

Byte hiByte( Word int_16 ) {           // take high byte of 16-bit word
    union_16 u;
    u.int_16 = int_16;
    return u.b.byte_2;       // Intel order
}

Byte loByte( Word int_16 ) {           // take low byte of 16-bit word
    union_16 u;
    u.int_16 = int_16;
    return u.b.byte_1;       // Intel order
}
