
/****************************************************************************
 *  Filename:     ni_callb.c
 *
 *  Description:  This file contains handler functions for network
 *                management messages coming from the network, or from
 *                the network interface node, to the host.
 *
 *  Copyright (c) 1994 Echelon Corporation.  All Rights Reserved.
 *
 ***************************************************************************/

#include "ni_msg.h"
#include "ni_callb.h"
#include "ni_mgmt.h"

// The following table, callbackTable[], defines the callback functions
//    that are called for specific incoming message types.

static NiCallbackData callbackTable[] =
{
/* procedure:                NM code                 callback type */
   { NiHandleAppMsgs,        0                 }, // APP_MSG_CALLBACK
   { NiHandleNvPoll,         0                 }, // NV_POLL_CALLBACK
   { NiHandleNvUpdate,       0                 }, // NV_UPDATE_CALLBACK

   { NiHandleWink,           NM_wink           }, // WINK_CALLBACK
   { NiHandleQuerySnvt,      NM_query_SNVT     }, // QUERY_SNVT_CALLBACK
   { NiHandleNvFetch,        NM_NV_fetch       }, // NV_FETCH_CALLBACK
   { NiHandleQueryNvConfig,  NM_query_nv_cnfg  }, // QUERY_NV_CNFG_CALLBACK
   { NiHandleUpdateNvConfig, NM_update_nv_cnfg }, // UPDATE_NV_CNFG_CALLBACK
   { NiHandleSetNodeMode,    NM_set_node_mode  }, // SET_NODE_MODE_CALLBACK
   { NiHandleServicePinMsg,  NM_service_pin    }, // SERVICE_PIN_CALLBACK

   { NiHandleInvalidResp,    0                 }, // INVALID_RESP_CALLBACK
   { NiHandleUplinkCmd,      0                 }  // UPLINK_CMD_CALLBACK
};

/***************************************************************************
   Function:  NiInvokeCallback()

   Desc:      Function to call a particular callback with an
              incoming message.  The callbackTable[] is used to determine
              the function to call, based on the CallbackType.

   ************************************************************************/

Bool NiInvokeCallback( CallbackType type, ExpAppBuffer * pMsgIn ) {
    NiCallbackProc callback;

    if( type >= NUM_CALLBACKS ) return FALSE;
      // Get the callback function from the table
    callback = callbackTable[ type ].proc;
    if( ! callback ) return FALSE;
    return callback( pMsgIn );
}

/***************************************************************************
   Function:  InvokeNmCallback()

   Desc:      This internal routine determines which network management
              callback function to call.  The callbackTable[] is used to
              determine the function to call based on the network management
              code.

   ************************************************************************/

static Bool InvokeNmCallback( Byte nmCommand, ExpAppBuffer * pMsgIn ) {
    NiCallbackData * pCallback = callbackTable;
    NiCallbackProc   callback;
    CallbackType     type;

   /* Loop through the callback table, looking to see if the network
      management command matches any entries in the table.  If so, invoke
      that callback: */

    for( type = 0; type < NUM_CALLBACKS; type++, pCallback++ ) {
        if( pCallback->nmCommand == nmCommand ) {
            callback = pCallback->proc;
            if( ! callback ) return FALSE;
            return callback( pMsgIn );
        }
    }
    return FALSE;
}

/**************************************************************************/

Bool NiCommandCallback( ExpAppBuffer * pMsgIn ) {

    if( pMsgIn->data.exp.code <= NA_last_msg_code )   // explicit message
        return NiInvokeCallback( APP_MSG_CALLBACK, pMsgIn );

    if( pMsgIn->data.unv.must_be_one ) {  // incoming NV message
            // If a REQUEST, it is a poll, otherwise it is an update

            if( pMsgIn->msg_hdr.exp.st == REQUEST )
               return NiInvokeCallback( NV_POLL_CALLBACK, pMsgIn );
            else
               return NiInvokeCallback( NV_UPDATE_CALLBACK, pMsgIn );
     }
              // Check for network management messages
     return InvokeNmCallback( pMsgIn->data.exp.code, pMsgIn );
}
