/* NI_MSG.C -- LONWORKS(R) host application message handler.
 * Copyright (c) 1993, 1994 by Echelon Corporation.
 * All Rights Reserved.
 *
 * Date last modified: 11/17/94
 *
 * Example LONWORKS network interface message send/receive functions.
 * Functions to initialize the network interface are also included.
 * These are general-purpose functions that may be used with any
 * host application.
 */

/****************************** Include Files *******************************/

#include "ldvintfc.h"
#include "ni_msg.h"
#include "hauif.h"
#include "ni_callb.h"
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <time.h>
//#include <dos.h>

/*
 ****************************************************************************
 * Default message buffers.
 * Assume explicit addressing is enabled in network interface
 ****************************************************************************
 */

ExpAppBuffer  msgIn;               // Incoming message buffer
ExpAppBuffer  msgOut;              // Outgoing message buffer

#define  INP_HDR     msgIn.msg_hdr.exp     // shorthand
#define  OUT_HDR     msgOut.msg_hdr.exp

/*
 ****************************************************************************
 * Structure to store incoming messages.
 ****************************************************************************
 */

struct msgListStruct {
    struct msgListStruct * nextMsg; // linked list pointer
    ExpAppBuffer           Msg;
};                                  // linked list of pending messages

typedef struct msgListStruct * pMsgList;

pMsgList pendingList  = NULL;       // List of pending incoming msgs
pMsgList responseList = NULL;       // List of subsequent responses

/*
 ****************************************************************************
 * Table to store priority of incoming request messages.
 ****************************************************************************
 */

#define NUM_RCV_TRANSACTIONS 16
Bool    requestPriority[ NUM_RCV_TRANSACTIONS ];

/***************************** Prototypes ******************************/

void initResponseList( void );
void initPendingList( void );
void savePending ( Byte msgLength );
void saveResponse( Byte msgLength );

NICode getMsg( Bool wait );
NICode putMsg( void );

typedef enum {          // internal enumeration for addressing mode
    local,
    implicit,
    explicit
} AddrMode;

void msgHdrInit( Byte msgSize, ServiceType service, Bool priority,
         Bool auth, AddrMode mode, Bits msgTag );

void msgDisplay( ExpAppBuffer * msgPtr );
void handleDriverError( const char * str, LDVCode ldvError );

/***************************** Local Variables ******************************/

const char * currentDeviceName; // Currently open network driver name
LNI niHandle = -1;              // Network driver handle

/*
 ****************************************************************************
 *
 *      NiSendMsgWait( ) - send a message and wait for completion
 *
 ****************************************************************************
 */

NICode NiSendMsgWait(
    ServiceType       service,        // ACKD, UNACKD_RPT, UNACKD, REQUEST
    const SendAddrDtl * outAddr,      // address of outgoing message
    const MsgData     * outData,      // data of outgoing message
    Byte                outLength,    // length of outgoing message
    Bool                priority,     // outgoing message priority
    Bool                outAuth,      // outgoing message authenticated
    ComplType         * completion,   // MSG_SUCCEEDS or MSG_FAILS
    int               * numResponses, // number of received responses
    RespAddrDtl       * inAddr,       // address of first response
    MsgData           * inData,       // data of first response
    Byte              * inLength,     // length of first response
    Bits                refID
) {
    AddrMode    mode;
    NICode      niError;
    Bits        inTag;
    Byte        msgLength;
    int         responseCount = 0;
    int         cmplCode;

    if( outAddr != NULL ) {              // address supplied
        if( outAddr->ua.type == LOCAL )
            mode = local;                // addressed to network interface
        else {
            mode = explicit;             // explicitly addressed
            msgOut.addr.snd = * outAddr; // copy destination address
        }
    } else mode = implicit;              // implicitly addressed

    // Copy outgoing data into msgOut

    memcpy( & msgOut.data.exp.code, outData, outLength );

    msgHdrInit( outLength,    // msg size
                service,      // service type
                priority,     // priority
                outAuth,      // authenticated
                mode,         // addressing mode
                refID );      // message tag

    niError = putMsg( );                    // send message
    if( niError != NI_OK ) return niError;

    if( numResponses ) * numResponses = 0;  // clear count

    if( ( mode == local ) && service != REQUEST ) {
                            // local, not request/response, all done
        if( completion ) * completion = MSG_SUCCEEDS;
        return NI_OK;
    }

    initResponseList( );    // clear list of responses
    for( ;; ) {             // look for completion events and responses
        niError = getMsg( TRUE );

        if( niError == NI_UPLINK_CMD ) {
                 // uplink network interface command
            ( void ) NiInvokeCallback( UPLINK_CMD_CALLBACK, & msgIn );
            continue;       // keep looking for responses and completion
        }
        if( niError != NI_OK ) return niError;

        msgLength = INP_HDR.length;
        inTag = INP_HDR.tag;

        if( msgIn.ni_hdr.q.queue == niINCOMING ) {   // an incoming message
            savePending ( msgLength );
            continue;       // keep looking for responses
        }

        if( INP_HDR.response ) {     // a response
            if( inTag != refID ) {  // not our tag
                NiInvokeCallback( INVALID_RESP_CALLBACK, & msgIn );
                continue;           // keep looking for completion
            }
                                    // a response to our tag
            if( ! responseCount ) { // first response, copy it
                if( inAddr ) * inAddr = msgIn.addr.rsp;
                if( inData ) memcpy( inData, & msgIn.data, msgLength );
                if( inLength ) * inLength = msgLength;
            } else saveResponse( msgLength );
                                    // subsequent response to our tag

            responseCount++;        // got another one

            if( mode != local ) continue; // keep looking for completion

          // no completion events for local msgs, only one response

            if( completion ) * completion = MSG_SUCCEEDS;
            if( numResponses ) * numResponses = responseCount;
            return NI_OK;         // This transaction is done
        }
        // here if incoming is not a response

        cmplCode = INP_HDR.cmpl_code;
        switch( cmplCode ) {
        case MSG_FAILS:                // found the completion event
        case MSG_SUCCEEDS:
            if( inTag != refID ) {    // not our tag
                NiInvokeCallback( INVALID_RESP_CALLBACK, & msgIn );
                return NI_INTERNAL_ERR;
            }
            if( completion ) * completion = cmplCode;
            if( numResponses ) * numResponses = responseCount;
            return NI_OK;               // this transaction is done

        default:                        // invalid completion code
            NiInvokeCallback( INVALID_RESP_CALLBACK, & msgIn );
            return NI_INTERNAL_ERR;
        }               // end switch( cmplCode )
    }                   // end for( ;; )
}

/*
 ****************************************************************************
 *
 *      NiGetNextResponse( ) - get subsequent responses to a request message
 *
 ****************************************************************************
 */

NICode NiGetNextResponse(         // get subsequent responses here
    RespAddrDtl     * inAddr,
    MsgData         * inData,
    Byte            * inLength
) {
    Byte         msgLength;
    pMsgList     nextResponse;

    if( responseList == NULL ) return NI_NO_RESPONSES;

    msgLength = responseList->Msg.msg_hdr.exp.length;     // get data
    if( inAddr ) * inAddr = responseList->Msg.addr.rsp;
    if( inData ) memcpy( inData, & responseList->Msg.data, msgLength );
    if( inLength ) * inLength = msgLength;

    nextResponse = responseList->nextMsg;       // unhook this response
    free( responseList );
    responseList = nextResponse;
    return NI_OK;
}

/*
 ****************************************************************************
 *
 *      processMsg - function to receive and dispatch a message
 *
 ****************************************************************************
 */

Bool processMsg( void ) {          // return TRUE if something was printed
    pMsgList      nextPending;
    NICode        niError;
    Bool          retFlag;

    if( pendingList != NULL ) {    // got one squirreled away

        if( pendingList->Msg.msg_hdr.exp.st == REQUEST )
            requestPriority[ pendingList->Msg.msg_hdr.exp.tag ]
                = pendingList->Msg.msg_hdr.exp.priority;
                     // save priority for response

        retFlag = NiCommandCallback( & pendingList->Msg );

        nextPending = pendingList->nextMsg;  // unhook pending message
        free( pendingList );
        pendingList = nextPending;
        return retFlag;
    }

    // nothing on the pending list, see if there is an incoming message

    niError = getMsg( FALSE );
    switch( niError ) {
    case NI_TIMEOUT:            // no message available
        return FALSE;
    case NI_OK:                 // message received
        break;
    case NI_UPLINK_CMD:         // uplink network interface command
        return NiInvokeCallback( UPLINK_CMD_CALLBACK, & msgIn );
    default:
        NiErrorDisplay( "Error receiving message: ", niError );
        return TRUE;
    }
    if( msgIn.ni_hdr.q.queue != niINCOMING )   // an incoming message
          return NiInvokeCallback( INVALID_RESP_CALLBACK, & msgIn );

    if( INP_HDR.st == REQUEST )       // save priority for a response
        requestPriority[ INP_HDR.tag ] = INP_HDR.priority;

    return NiCommandCallback( & msgIn );     // invoke handler
}

/*
 ****************************************************************************
 *
 *      NiSendResponse( ) - send a response to a request message
 *
 ****************************************************************************
 */

NICode NiSendResponse(         // send response to last received request
    MsgData       * outData,   // data for outgoing response
    Byte            outLength, // length of outgoing response
    Bits            refID
) {
    Bool            priority;

    priority = requestPriority[ refID ];    // priority of this transaction
    msgOut.ni_hdr.q.queue        = priority ? niNTQ_P : niNTQ;
    msgOut.ni_hdr.q.q_cmd        = niCOMM;
    msgOut.ni_hdr.q.length       = sizeof( ExpMsgHdr )
                                 + sizeof( ExplicitAddr) + outLength;
    OUT_HDR.tag       = refID;
    OUT_HDR.auth      = 0;
    OUT_HDR.response  = 1;       // Make it a response/
    OUT_HDR.st        = REQUEST;
    OUT_HDR.pool      = 0;       // Must be zero
    OUT_HDR.alt_path  = 0;       // Use default path
    OUT_HDR.addr_mode = 0;       // IMplicit addressing
    OUT_HDR.cmpl_code = MSG_NOT_COMPL; // Zero
    OUT_HDR.path      = 0;       // Use primary path
    OUT_HDR.priority  = priority;
    OUT_HDR.length    = outLength;

    memcpy( & msgOut.data.exp, outData, outLength );    // copy data
    return putMsg( );                                   // send message
};

/************************************************************************/

static void initResponseList( void ) {   // initialize list of responses

    pMsgList    nextResponse;

    while( responseList != NULL ) {
        nextResponse = responseList->nextMsg;
        free( responseList );
        responseList = nextResponse;
   }
}
/************************************************************************/

static void initPendingList( void ) {   // initialize list of pending messages

    pMsgList    nextPending;

    while( pendingList != NULL ) {
        nextPending = pendingList->nextMsg;
        free( pendingList );
        pendingList = nextPending;
   }
}
/************************************************************************/

static void saveResponse( Byte msgLength ) {
     // come here to save any responses beyond the first for group
     // addressed request/response service

    pMsgList    nextResponse;
    Byte        bufSize;

    bufSize =  sizeof( pMsgList ) + sizeof( NI_Hdr )
             + sizeof( MsgHdr )   + sizeof( ExplicitAddr )
             + msgLength;

    nextResponse = ( pMsgList ) malloc( bufSize );
        // allocate a buffer for the incoming message

    if( ! nextResponse ) {
        printf( "Out of memory in saveResponse\n" );
        exit( 1 );      // Unfriendly recovery
    }
    nextResponse->nextMsg = responseList;        // link chain
    memcpy( & nextResponse->Msg, & msgIn, bufSize );      // copy message
    responseList = nextResponse;
}

/************************************************************************/

static void savePending( Byte msgLength ) {
    // come here to squirrel away an incoming message that is not part
    // of the current transaction
    pMsgList    nextPending;
    Byte        bufSize;

    bufSize =  sizeof( pMsgList ) + sizeof( NI_Hdr )
             + sizeof( MsgHdr )   + sizeof( ExplicitAddr )
             + msgLength;

    nextPending = ( pMsgList ) malloc( bufSize );
        // allocate a buffer for the incoming message

    if( ! nextPending ) {
        printf( "Out of memory in savePending\n" );
        exit( 1 );      // Unfriendly recovery
    }
    nextPending->nextMsg = pendingList;    // link chain
    memcpy( & nextPending->Msg, & msgIn, bufSize );     // copy message
    pendingList = nextPending;
}

/*
 ****************************************************************************
 *
 * getMsg().  Receive a message in msgIn.  Return when a message
 * is received, or if the wait timer times out
 *
 ****************************************************************************
 */

static NICode getMsg( Bool wait ) {
           // wait : should this call wait for input?

    LDVCode ldvError;
    time_t  deadline;

    if( wait ) deadline = time( NULL ) + niWAIT_TIME;

    // Loop until the network interface provides a message, or timeout
    for( ;; ) {
        ldvError = ldv_read( niHandle, & msgIn, sizeof( msgIn ) );
                // Check for network interface input

        switch( ldvError ) {

        case LDV_OK:
            if( verboseFlag ) {
                printf( "\nReceived a LONTALK message from the interface\n" );
                msgDisplay( & msgIn );
            }
            if( msgIn.ni_hdr.q.q_cmd != niCOMM ) return NI_UPLINK_CMD;
            return NI_OK;

        case LDV_NO_MSG_AVAIL:
        case LDV_DEVICE_BUSY:
            if( ! wait || ( time( NULL ) > deadline ) ) return NI_TIMEOUT;
            continue;

        default:
            handleDriverError( "Driver error on read", ldvError );
            return NI_DRIVER_ERROR;
        }                           // end switch
    }                               // end for
}

/*
 ****************************************************************************
 *
 * putMsg().  Send the message assembled in msgOut.
 *
 ****************************************************************************
 */

static NICode putMsg( void ) {
    LDVCode ldvError;
    time_t  deadline;

    deadline =  time( NULL ) + niWAIT_TIME;

        // Loop until the network interface accepts the output, or timeout.
    if( verboseFlag ) {
        printf( "\nSent a LONTALK message to the interface\n" );
        msgDisplay( & msgOut );
    }

    for( ;; ) {
        ldvError = ldv_write( niHandle, & msgOut, sizeof( msgOut ) );
            // Attempt to write to the network interface.

        switch( ldvError ) {

        case LDV_OK:
           return NI_OK;

        case LDV_NO_BUFF_AVAIL:
        case LDV_DEVICE_BUSY:
            if( time( NULL ) > deadline ) return NI_TIMEOUT;
            continue;

        default:
            handleDriverError( "Driver error on write", ldvError );
            return NI_DRIVER_ERROR;
        }                           // end switch
    }                               // end for
}

/*
 ****************************************************************************
 *
 * msgHdrInit().  Initialize an outgoing application buffer
 * for a network interface using explicit addressing format
 *
 ****************************************************************************
 */

static void msgHdrInit( Byte msgSize, ServiceType service, Bool priority,
       Bool auth, AddrMode mode, Bits msgTag )
{

    const static NI_Queue queue[ 2 ][ 2 ] =         // define MIP queue
        { { niTQ, niTQ_P }, { niNTQ, niNTQ_P } };

    msgOut.ni_hdr.q.queue  = queue[ service == UNACKD ][ priority ];
                // Get correct queue

    msgOut.ni_hdr.q.q_cmd    = ( mode == local ) ? niNETMGMT : niCOMM;
             // Data transfer for network interface or network

    msgOut.ni_hdr.q.length   = sizeof( ExpMsgHdr ) + sizeof( ExplicitAddr )
                             + msgSize;     // Header size

    OUT_HDR.tag       = msgTag;
    OUT_HDR.auth      = auth;
    OUT_HDR.st        = service;
    OUT_HDR.msg_type  = 0;              // MIP doesn't process NVs
    OUT_HDR.response  = 0;              // Not a response message
    OUT_HDR.pool      = 0;              // Must be zero
    OUT_HDR.alt_path  = 0;              // Use default path
    OUT_HDR.addr_mode = ( mode == explicit ); // Addressing mode
    OUT_HDR.cmpl_code = MSG_NOT_COMPL;  // Zero
    OUT_HDR.path      = 0;              // Use primary path
    OUT_HDR.priority  = priority;
    OUT_HDR.length    = msgSize;        // Message size
}

/*
 ****************************************************************************
 *
 *      msgDisplay().  Display contents of a network interface message.
 *
 ****************************************************************************
 */

static void msgDisplay( ExpAppBuffer * msgPtr ) {
    int    length;
    Byte * pData;

    static const char * cmdStr[ 16 ] = {
        "NULL", "COMM", "NETMGMT", "TIMEOUT", "CRC", "RESET", "FLUSH_C",
        "ONLINE", "OFFLINE", "FLUSH", "FLUSH_IGN", "SLEEP", "(N)ACK",
        "0xD", "0xE", "DRV_CMD" };

    static const char * queueStr[ 16 ] = {
        "0x0", "0x1", "TQ", "TQ_P", "NTQ", "NTQ_P", "RESPONSE", "0x7",
        "INCOMING", "0x9", "0xA", "0xB", "0xC", "0xD", "0xE", "0xF" };

    static const char * complStr[ 4 ] = {
        "<none>", "MSG_SUCCEEDS", "MSG_FAILS", "cmpl=3" };

    // Display headers.
    length = msgPtr->ni_hdr.q.length;
    printf( " Network intfc header: command = ni%s, queue = ni%s, "
              "total length = %d\n", cmdStr[ msgPtr->ni_hdr.q.q_cmd ],
              queueStr[ msgPtr->ni_hdr.q.queue ], length );

    if( ! length ) return;      // nothing to print

    printf( " Message header: tag = %d, service = %s, response = %s, "
              "data length = %d\n", msgPtr->msg_hdr.exp.tag,
              svcTable[ msgPtr->msg_hdr.exp.st ],
              msgPtr->msg_hdr.exp.response ? "yes": "no",
              msgPtr->msg_hdr.exp.length );

    printf( " Explicit address mode = %s, completion code = %s\n",
              msgPtr->msg_hdr.exp.addr_mode ? "yes" : "no",
              complStr[ msgPtr->msg_hdr.exp.cmpl_code ] );

    // Display address field
    printf( " Address field (values in hex):\n" );
    printArray( & msgPtr->addr, sizeof( ExplicitAddr ), "%02X " );
    printf( "\n" );

    // Display data field.

    length = min( length - sizeof( MsgHdr ) - sizeof( ExplicitAddr ),
                  msgPtr->msg_hdr.exp.length );
                                            // length from Msg header
    if( ! length ) return;      // nothing to print

    printf( " Data field (values in hex):\n" );
    pData = ( Byte * ) & msgPtr->data;

    while( length > 0 ) {
        printArray( pData, min( 16, length ), "%02X " );
        printf( "\n" );
        length -= 16;
        pData += 16;
    }
}

/*
 ****************************************************************************
 *
 * NiHandleError - report errors from a LonTalk message. Checks niError
 * against NI_OK, completion against MSG_SUCCEEDS, and checks response_code
 * for a successful network management response bit
 * (use NO_CHECK if it's not a network management message, or not a response)
 * Returns FALSE if everything is OK, TRUE if some error occurred.
 *
 ****************************************************************************
 */

Bool NiHandleError( NICode niError,
                           ComplType    completion,
                           Byte         responseCode,
                           const char * msgName ) {

    if( niError != NI_OK ) {
        printf( "Network interface error: " );
        NiErrorDisplay( msgName, niError );
        return TRUE;
    }
    if( completion == MSG_FAILS ) {
        printf( "Message failed completion: %s\n", msgName );
        return TRUE;
    }
    if( !( responseCode & NO_CHECK ) ) {
        printf( "Response error: %s\n", msgName );
        return TRUE;
    }
    return FALSE;
}

/*
 ****************************************************************************
 *
 * NiInit().  Initialize the network interface.  Called on
 * program startup and when network interface errors occur.  Network
 * interface errors can be caused by cycling power on an SLTA with
 * autobaud selected while this program is running.  Reopening the network
 * driver will reinitialize the bit rate on the SLTA interface.
 *
 ****************************************************************************
 */

NICode NiInit( const char * deviceName ) {
    NICode  niError;
    LDVCode ldvError;

    if( niHandle >= 0 ) ldv_close( niHandle );
        // Network interface has already been opened, close it first

    // Open network interface
    ldvError = ldv_open( deviceName, & niHandle );
    if( ldvError != LDV_OK ) {
        ldvErrorDisplay( "Driver error on open", ldvError );
        return NI_DRIVER_NOT_INIT;
    }

    // Empty out any waiting uplink messages

    niError = NiSendImmediate( niRESET );
    if( niError != NI_OK ) return NI_DRIVER_NOT_RESET;
    niError = getMsg( TRUE );       // wait for the uplink reset
    if( niError != NI_UPLINK_CMD ) return NI_DRIVER_NOT_RESET;

    currentDeviceName = deviceName;
    return NI_OK;
}

/*
 ****************************************************************************
 *
 * NiClose().  Close network interface.
 *
 ****************************************************************************
 */

void NiClose( void ) {
    ldv_close( niHandle );
    niHandle = -1;
}

/*
 ****************************************************************************
 *
 * NiSendImmediate( ).  Send an immediate network interface command
 *
 ****************************************************************************
 */

NICode NiSendImmediate( NI_NoQueueCmd command ) {
    msgOut.ni_hdr.noq.cmd = command;
    msgOut.ni_hdr.noq.length = 0;
    return putMsg( );
}

/*
 ****************************************************************************
 *
 * handleDriverError( ) - display error from driver, and try to recover
 *
 ****************************************************************************
 */

static void handleDriverError( const char * str, LDVCode ldvError ) {


    ldvErrorDisplay( str, ldvError );               // Print error message

    if ( NiInit( currentDeviceName ) != NI_OK )     // Try to re-open driver
        printf( "Could not reopen network driver\n" );
    else
        printf( "Reopened network driver\n" );
}

/*
 ****************************************************************************
 *
 * Network interface error strings.
 *
 ****************************************************************************
 */

static char * niErrorStrings[] = {
    "No error",
    "Device name not specified",
    "Could not open network driver",
    "Could not initialize network driver",
    "Could not reset network driver",
    "Network driver error",
    "No response messages left",
    "Reset failed",
    "Timeout",
    "Uplink command received",
    "Internal error detected",
    "File open error"
};

/*
 ****************************************************************************
 *
 * NiErrorDisplay().  Display a network interface error message.  Display
 * a user specified string first.
 *
 ****************************************************************************
 */

void NiErrorDisplay( const char * str, NICode niError ) {
    if( niError == NI_OK ) return;
    errorDisplay( str, niErrorStrings[ niError ] );
}
